from chroma_datasets.types import AddEmbedding, Datapoint, Dataset
from chroma_datasets.utils import transform_data, load_huggingface_dataset, to_chroma_schema
from typing import Optional, Union, Sequence, Dict, Mapping, List
from datasets import load_dataset


class Glue(Dataset):
    """
    This dataset contains the popular Glue Dataset.
    It is hardcoded to "ax" subdataset and the "test" split.

    Columns:
        - premise: The premise of the sentence pair.
        - hypothesis: The hypothesis of the sentence pair.
        - label: The label of the sentence pair.
        - idx: The index of the sentence pair.
    """
    hf_data = None

    @classmethod
    def load_data(cls):
        cls.hf_data = load_dataset("glue", "ax", split="test")

    @classmethod
    def raw_text(cls) -> str:
        if cls.hf_data is None:
            cls.load_data()
        return "\n".join(cls.hf_data["premise"])
    
    @classmethod
    def chunked(cls) -> List[Datapoint]:
        if cls.hf_data is None:
            cls.load_data()

        mapping = {
            "id": lambda row: str(row["idx"]),
            "metadata": lambda row: {"premise": row["premise"], "hypothesis": row["hypothesis"], "label": row["label"]},
            "document": "premise"
        }
        return transform_data(cls.hf_data, mapping)
    
    @classmethod
    def to_chroma(cls) -> AddEmbedding:
        return to_chroma_schema(cls.chunked())



from chroma_datasets.types import AddEmbedding, Datapoint, Dataset
from chroma_datasets.utils import transform_data, load_huggingface_dataset, to_chroma_schema
from langchain.text_splitter import CharacterTextSplitter
from typing import Optional, Union, Sequence, Dict, Mapping, List


class PaulGrahamEssay(Dataset):
    """
    http://www.paulgraham.com/worked.html

    Columns:
        - id: unique identifier for each chunk
        - document: the text of the chunk
        - embedding: the embedding of the chunk (OpenAI-ada-002)
        - metadata: metadata about the chunk
    """
    hf_data = None
    embedding_function = "OpenAIEmbeddingFunction" # name of embedding function inside Chroma
    embedding_function_instructions = """
        from chromadb.utils import embedding_functions
        openai_ef = embedding_functions.OpenAIEmbeddingFunction(
            api_key="YOUR_API_KEY",
            model_name="text-embedding-ada-002"
        )
    """

    @classmethod
    def load_data(cls):
        cls.hf_data = load_huggingface_dataset(
            "chromadb/paul_graham_essay",
            split_name="data"
        )

    @classmethod
    def raw_text(cls) -> str:
        if cls.hf_data is None:
            cls.load_data()
        return "\n".join(cls.hf_data["document"])
    
    @classmethod
    def chunked(cls) -> List[Datapoint]:
        if cls.hf_data is None:
            cls.load_data()
        return cls.hf_data
    
    @classmethod
    def to_chroma(cls) -> AddEmbedding:
        return to_chroma_schema(cls.chunked())



from chroma_datasets.types import AddEmbedding, Datapoint, Dataset
from chroma_datasets.utils import transform_data, load_huggingface_dataset, to_chroma_schema
from langchain.text_splitter import CharacterTextSplitter
from typing import Optional, Union, Sequence, Dict, Mapping, List
import uuid

class SciQ(Dataset):
    """
    https://huggingface.co/datasets/sciq

    Columns:
        - question: str
        - correct_answer: str
        - support: str
        - distractor1: str
        - distractor2: str
        - distractor3: str
    """
    hf_data = None
    embedding_function = None

    @classmethod
    def load_data(cls):
        cls.hf_data = load_huggingface_dataset(
            "sciq",
            split_name="test"
        )

    @classmethod
    def raw_text(cls) -> str:
        """
            Doesn't make sense for this dataset
        """
        raise NotImplementedError
    
    @classmethod
    def chunked(cls) -> List[Datapoint]:
        """
            The dataset is already chunked
        """
        if cls.hf_data is None:
            cls.load_data()

        dataset_rows = []
        for row in cls.hf_data:

            # add questions
            dataset_rows.append({
                "id": str(uuid.uuid4()),
                "document": row["question"],
                "metadata": {  
                    "type": "question",
                }
            })

            # add supporting evidence
            dataset_rows.append({
                "id": str(uuid.uuid4()),
                "document": row["support"],
                "metadata": {
                    "type": "supporting_evidence"
                }
            })

        return dataset_rows
    
    @classmethod
    def to_chroma(cls) -> AddEmbedding:
        return to_chroma_schema(cls.chunked())



__version__ = "0.1.3"

from .datasets.state_of_the_union import StateOfTheUnion
from .datasets.glue import Glue
from .datasets.paul_graham_essay import PaulGrahamEssay
from .datasets.sciq import SciQ



from abc import ABC, abstractmethod
from typing import Optional, Union, Sequence, Dict, Mapping, List, Any
from typing_extensions import TypedDict


Metadata = Mapping[str, Union[str, int, float]]
Vector = Union[Sequence[float], Sequence[int]]


class Datapoint(TypedDict):
    id: Optional[str]
    metadata: Optional[Metadata]
    document: Optional[str] 
    embedding: Optional[Vector]


class AddEmbedding(TypedDict): 
    embeddings: Optional[List[Any]] 
    metadatas: Optional[List[Dict[Any, Any]]] 
    documents: Optional[List[str]] 
    ids: List[str]


class Dataset(ABC):
    """
        Abstract class for a dataset

        All datasets should inherit from this class

        Properties:
            hf_data: the raw data from huggingface
            embedding_function: the embedding function used to generate the embeddings
            embeddingFunctionInstructions: tell the user how to set up the embedding function
    """
    hf_data: Any
    embedding_function: str
    embedding_function_instructions: str

    @classmethod
    @abstractmethod
    def raw_text(cls) -> str:
        """
            Returns the dataset as one long string.
            Useful if you want to chunk it yourself.
            Note: you will reset chunk-specific metadata, etc if it exists.
        """
        pass

    @classmethod
    @abstractmethod
    def chunked(cls) -> List[Datapoint]:
        """
            Returns the dataset with a default chunking scheme
        """
        pass

    @classmethod
    @abstractmethod
    def to_chroma(cls) -> AddEmbedding:
        """
            Returns the dataset in a format that can be used by Chroma
        """
        pass




from datasets import load_dataset, DatasetInfo, Dataset, load_from_disk
from typing import Optional, Union, Sequence, Dict, Mapping, List
from chroma_datasets.types import AddEmbedding, Datapoint


def load_huggingface_dataset(dataset_name, split_name=None):
    """
    Loads a dataset from Hugging Face datasets library.

    This is a very minimal wrapper, but we are including it to make it easier to follow the logic
    for importing future kinds of datasets.

    Args:
        dataset_name (str): Name of the dataset to load.
        split_name (str, optional): Name of the split to load (e.g., "train", "test", "validation"). 
                                   If not provided, the entire dataset will be loaded.

    Returns:
        dataset (datasets.Dataset): The loaded dataset.
    """
    dataset = load_dataset(dataset_name, split=split_name)
    return dataset


def transform_data(dataset, mapping):
    """
    Transforms a dataset using a mapping.

    Optional utility function that can be used to transform a dataset into a format that Chroma can understand.

    See `glue.py` for an example
    """
    # Create a new list for the transformed data
    transformed_data = []

    for i, row in enumerate(dataset):
        # Create a new dictionary for the transformed row
        transformed_row = {}

        # Process each item in the mapping
        for key, value in mapping.items():
            # If the value is a string, perform a direct mapping
            if isinstance(value, str):
                transformed_row[key] = row[value]

            # If the value is a function, use it to transform the data
            elif callable(value):
                transformed_row[key] = value(row)

            else:
                raise ValueError(f"Invalid mapping for {key}: {value}")

        # Add the transformed row to the list
        transformed_data.append(transformed_row)

    # Return the transformed data
    return transformed_data


def to_chroma_schema(datapoints: List[Datapoint]) -> AddEmbedding:
    """
    Converts a list of datapoints into the AddEmbedding schema.
    """
    embeddings = []
    metadatas = []
    documents = []
    ids = []

    for datapoint in datapoints:

        # datapoint is an object, check to see if it has a key called "embedding"
        # if it does, add it to the add_embedding object
        # if it doesn't, do nothing
        if "embedding" in datapoint:
            embeddings.append(datapoint["embedding"])
        if "metadata" in datapoint:
            metadatas.append(datapoint["metadata"])
        if "document" in datapoint:
            documents.append(datapoint["document"])
        if "id" in datapoint:
            ids.append(datapoint["id"])

    return {
        "ids": ids if len(ids) > 0 else None,
        "metadatas": metadatas if len(metadatas) > 0 else None,
        "documents": documents if len(documents) > 0 else None,
        "embeddings": embeddings if len(embeddings) > 0 else None,
    }


def import_into_chroma(chroma_client, dataset, collection_name=None, embedding_function=None):
    """
    Imports a dataset into Chroma.

    Args:
        chroma_client (ChromaClient): The ChromaClient to use.
        collection_name (str): The name of the collection to load the dataset into.
        dataset (AddEmbedding): The dataset to load.
        embedding_function (Optional[Callable[[str], np.ndarray]]): A function that takes a string and returns an embedding.
    """
    # if chromadb is not install, raise an error
    try:
        import chromadb
        from chromadb.utils import embedding_functions
    except ImportError:
        raise ImportError("Please install chromadb to use this function. `pip install chromadb`")

    ef = None

    if dataset.embedding_function is not None:
        if embedding_function is None:
            error_msg = "See documentation"
            if dataset.embedding_function_instructions is not None:
                error_msg = dataset.embedding_function_instructions

            raise ValueError(f"""
                             Dataset requires embedding function: {dataset.embedding_function}.

                             {error_msg}
                             """)

        if embedding_function.__class__.__name__ != dataset.embedding_function:
            raise ValueError(f"Please use {dataset.embedding_function} as the embedding function for this dataset. You passed {embedding_function.__class__.__name__}")

    if embedding_function is not None:
        ef = embedding_function

    # if collection_name is None, get the name from the dataset type 
    if collection_name is None:
        collection_name = dataset.__name__

    if ef is None:
        ef = embedding_functions.DefaultEmbeddingFunction()

    collection = chroma_client.create_collection(
        collection_name,
        embedding_function=ef
        )

    mapped_data = dataset.to_chroma()

    collection.add(
        ids=mapped_data["ids"],
        metadatas=mapped_data["metadatas"],
        documents=mapped_data["documents"],
        embeddings=mapped_data["embeddings"],
    )

    print(f"Loaded {len(mapped_data['ids'])} documents into the collection named: {collection_name}")

    return collection


def import_chroma_exported_hf_dataset_from_disk(chroma_client, path, collection_name, embedding_function=None):
    dataset = load_from_disk(path)
    return import_chroma_exported_hf_dataset(chroma_client, dataset, collection_name, embedding_function=embedding_function)


def import_chroma_exported_hf_dataset(chroma_client, dataset, collection_name, embedding_function=None):
    """
        Imports a dataset that was exported from Chroma into a Hugging Face dataset.
        This will only work with data exported from Chroma and is not a generic utility function.
    """
    try:
        import chromadb
        from chromadb.utils import embedding_functions
    except ImportError:
        raise ImportError("Please install chromadb to use this function. `pip install chromadb`")

    if embedding_function is None:
        print("Caution: No embedding function provided. Using the default embedding function.")
        embedding_function = embedding_functions.DefaultEmbeddingFunction()

    if collection_name is None:
        raise ValueError("Please provide a collection name.")

    # dataset = load_from_disk(path)
    mapped_data = to_chroma_schema(dataset)

    collection = chroma_client.create_collection(
        collection_name,
        embedding_function=embedding_function
        )

    collection.add(
        ids=mapped_data["ids"],
        metadatas=mapped_data["metadatas"],
        documents=mapped_data["documents"],
        embeddings=mapped_data["embeddings"],
    )

    print(f"Loaded {len(mapped_data['ids'])} documents into the collection named: {collection_name}")

    return collection


def export_collection_to_hf_dataset_to_disk(chroma_client, collection_name, path, license="MIT"):
    dataset = export_collection_to_hf_dataset(chroma_client, collection_name, license=license)
    dataset.save_to_disk(path)
    print(f"Saved dataset to {path}")


def export_collection_to_hf_dataset(chroma_client, collection_name, license="MIT"):
    """
    Exports a Chroma collection to a Hugging Face dataset.

    Args:
        chroma_client (ChromaClient): The ChromaClient to use.
        collection_name (str): The name of the collection to export.
        license (str): The license to use for the dataset. Default MIT
    """
    collection = chroma_client.get_collection(collection_name)

    # add pagination later
    all_data = collection.get(include=["documents", "embeddings", "metadatas"])
    
    row_data = []
    for id, embdding, metadata, document in zip(all_data["ids"], all_data["embeddings"], all_data["metadatas"], all_data["documents"]):
        row_data.append({"id": id, "embedding": embdding, "metadata": metadata, "document": document})

    dataset_info = DatasetInfo(
            license=license,
        )

    dataset = Dataset.from_list(
        row_data,
        info=dataset_info
    )

    return dataset


from chroma_datasets.types import AddEmbedding, Datapoint, Dataset
from chroma_datasets.utils import transform_data, load_huggingface_dataset, to_chroma_schema
from langchain.text_splitter import CharacterTextSplitter
from typing import Optional, Union, Sequence, Dict, Mapping, List


class StateOfTheUnion(Dataset):
    """
    This dataset contains the text of the 2022 State of the Union Address.

    Columns:
        - text: The text of the 2022 State of the Union address.
    """
    hf_data = None
    embedding_function = None

    @classmethod
    def load_data(cls):
        cls.hf_data = load_huggingface_dataset(
            "chromadb/state_of_the_union",
            split_name="data"
        )

    @classmethod
    def raw_text(cls) -> str:
        if cls.hf_data is None:
            cls.load_data()
        return "\n".join(cls.hf_data["text"])
    
    @classmethod
    def chunked(cls) -> List[Datapoint]:
        if cls.hf_data is None:
            cls.load_data()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0, separator="\n")
        split = text_splitter.split_text(cls.raw_text())

        # convert to array of Datapoint types
        split = [
            {
                "id": str(i),
                "document": chunk
            }
            for i, chunk in enumerate(split)
        ]

        return split
    
    @classmethod
    def to_chroma(cls) -> AddEmbedding:
        return to_chroma_schema(cls.chunked())
